/*
Copyright (C) 2006 Evan Teran
                   eteran@alum.rit.edu

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "version.h"
#include "DebuggerMain.h"
#include "Debugger.h"
#include "State.h"
#include "CommentServer.h"
#include "ArchProcessorInterface.h"
#include "DebuggerPluginInterface.h"
#include "DebugEventServer.h"
#include "ByteShiftArray.h"
#include "SymbolManager.h"
#include "DialogAttach.h"
#include "DialogMemoryRegions.h"
#include "Configuration.h"
#include "Expression.h"
#include "BinaryInfo.h"
#include "DebuggerCoreInterface.h"
#include "RecentFileManager.h"
#include "AnalyzerInterface.h"
#include "DialogPlugins.h"

#include <QDir>
#include <QToolButton>
#include <QMessageBox>
#include <QStringListModel>
#include <QFileDialog>
#include <QInputDialog>
#include <QSettings>
#include <QVector>
#include <QtDebug>

#include <sys/types.h>
#include <sys/stat.h>

#include <cstring>

//#define USE_SESSION

//------------------------------------------------------------------------------
// Name: DebuggerMain(QWidget *parent)
// Desc: constructor
//------------------------------------------------------------------------------
DebuggerMain::DebuggerMain(QWidget *parent) : QMainWindow(parent), 
		m_DebugEventServer(0), m_TTYProc(0), m_GUIState(TERMINATED), 
		m_StackViewLocked(false), m_StepWasRun(false), 
		m_ReenableBreakpointAddress(0) {

	setupUI();
	
	// register our DebugEvent type so it can be used in 
	// inter-thread signals
	qRegisterMetaType<DebugEvent>("DebugEvent");
		
	// setup the list model for instruction details list
	m_ListModel = new QStringListModel(this);
	ui.listView->setModel(m_ListModel);
	
	// setup the recent file manager
	m_RecentFileManager = new RecentFileManager;
	ui.action_Recent_Files->setMenu(m_RecentFileManager->createMenu());
	connect(m_RecentFileManager, SIGNAL(fileSelected(const QString &)), SLOT(doOpenFile(const QString &)));

	// make us the default event handler
	edb::v1::setDebugEventHandler(this);
		
	// we placed a view in the designer, so just set it here
	// this may get transitioned to heap allocated, we'll see
	m_StackViewInfo.view = ui.stackView;
	
	// enable the arch processor
	edb::v1::archProcessor().setupRegisterView(ui.registerList);

	// setup the comment server for the stack viewer
	m_CommentServer = new CommentServer;
	ui.stackView->setCommentServer(m_CommentServer);
		
	// default the working directory to ours
	m_WorkingDirectory = QDir().absolutePath();
}

//------------------------------------------------------------------------------
// Name: setupUI()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::setupUI() {
	// setup the global pointers as early as possible. 
	// NOTE:  this should never be changed after this point
	// NOTE:  this is important that this happens BEFORE any components which
	// read settings as it could end up being a memory leak (and therefore never
	// calling it's destructor which writes the settings to disk!).
	edb::v1::debuggerUI = this;

	ui.setupUi(this);

	// add toggles for the dock windows
	ui.menu_View->addAction(ui.registersDock->toggleViewAction());
	ui.menu_View->addAction(ui.dataDock->toggleViewAction());
	ui.menu_View->addAction(ui.stackDock->toggleViewAction());
	ui.menu_View->addAction(ui.toolBar->toggleViewAction());
	
	// make sure our widgets use custom context menus
	ui.registerList->setContextMenuPolicy(Qt::CustomContextMenu);
	ui.cpuView->setContextMenuPolicy(Qt::CustomContextMenu);
	ui.stackView->setContextMenuPolicy(Qt::CustomContextMenu);
	
	// set the listbox to about 4 lines
	const QFontMetrics &fm = ui.listView->fontMetrics();
	ui.listView->setFixedHeight(fm.height() * 4);
	
	// setup tabs
	setupTabButtons();
	
	// remove the one in the designer and put in our built one.
	// it's a bit ugly to do it this way, but the designer wont let me 
	// make a tabless entry..and the ones i create look slightly diff 
	// (probably lack of layout manager in mine...
	ui.tabWidget->removeTab(0);
	mnuDumpCreateTab();
	
	// apply any fonts we may have stored
	doApplyDefaultFonts();
	
	// apply the default setting for showing address seperators
	doApplyDefaultShowSeperator();
}

//------------------------------------------------------------------------------
// Name: ~DebuggerMain()
// Desc: destructor
//------------------------------------------------------------------------------
DebuggerMain::~DebuggerMain() {

	doDetachFromProcess(edb::v1::config().closeBehaviour == Configuration::Terminate);
	
	release(m_CommentServer);
	release(m_RecentFileManager);
	
	// remove the last tab
	mnuDumpDeleteTab();
	
	// kill our xterm and wait for it to die
	if(m_TTYProc != 0) {
		m_TTYProc->kill();
		m_TTYProc->waitForFinished(30000);
	}
}

//------------------------------------------------------------------------------
// Name: closeEvent(QCloseEvent *)
// Desc: triggered on main window close, saves window state
//------------------------------------------------------------------------------
void DebuggerMain::closeEvent(QCloseEvent *) {
	QSettings settings;
	const QByteArray state = saveState();
	settings.beginGroup("Window");
	settings.setValue("window.state", state);
	settings.setValue("window.width", width());
	settings.setValue("window.height", height());
	settings.setValue("window.stack.show_address.enabled", ui.stackView->showAddress());
	settings.setValue("window.stack.show_hex.enabled", ui.stackView->showHexDump());
	settings.setValue("window.stack.show_ascii.enabled", ui.stackView->showAsciiDump());
	settings.setValue("window.stack.show_comments.enabled", ui.stackView->showComments());
	settings.setValue("window.stack.row_width", ui.stackView->rowWidth());
	settings.setValue("window.stack.word_width", ui.stackView->wordWidth());
	settings.endGroup();
}

//------------------------------------------------------------------------------
// Name: showEvent(QShowEvent *)
// Desc: triggered on show, restores window state
//------------------------------------------------------------------------------
void DebuggerMain::showEvent(QShowEvent *) {
	QSettings settings;
	settings.beginGroup("Window");
	const QByteArray state = settings.value("window.state").value<QByteArray>();
	const int width = settings.value("window.width", -1).value<int>();
	const int height = settings.value("window.height", -1).value<int>();
	
	if(width != -1) {
		resize(width, size().height());
	}
	
	if(height != -1) {
		resize(size().width(), height);
	}
	
	ui.stackView->setShowAddress(settings.value("window.stack.show_address.enabled", true).value<bool>());
	ui.stackView->setShowHexDump(settings.value("window.stack.show_hex.enabled", true).value<bool>());
	ui.stackView->setShowAsciiDump(settings.value("window.stack.show_ascii.enabled", false).value<bool>());
	ui.stackView->setShowComments(settings.value("window.stack.show_comments.enabled", true).value<bool>());

	int rowWidth = settings.value("window.stack.row_width", 1).value<int>();
	int wordWidth = settings.value("window.stack.word_width", QT_POINTER_SIZE).value<int>();
	
	// normalize values
	if(wordWidth != 1 && wordWidth != 2 && wordWidth != 4 && wordWidth != 8) {
		wordWidth = QT_POINTER_SIZE;
	}
	
	if(rowWidth != 1 && rowWidth != 2 && rowWidth != 4 && rowWidth != 8 && rowWidth != 16) {
		rowWidth = 1;
	}
	
	ui.stackView->setRowWidth(rowWidth);
	ui.stackView->setWordWidth(wordWidth);
	
	settings.endGroup();
	restoreState(state);
}

//------------------------------------------------------------------------------
// Name: on_actionAbout_QT_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_actionAbout_QT_triggered() {
	QMessageBox::aboutQt(this, tr("About QT"));
}

//------------------------------------------------------------------------------
// Name: doApplyDefaultFonts()
// Desc: applies the confirations fonts to all neccessary widgets
//------------------------------------------------------------------------------
void DebuggerMain::doApplyDefaultFonts() {

	QFont font;
	const Configuration &config = edb::v1::config();

	// set some default fonts
	if(font.fromString(config.stackFont)) {
		ui.stackView->setFont(font);
	}
	
	if(font.fromString(config.registersFont)) {
		ui.registerList->setFont(font);
	}

	if(font.fromString(config.disassemblyFont)) {
		ui.cpuView->setFont(font);
	}
	
	if(font.fromString(config.dataFont)) {
		foreach(DataViewInfo *dataView, m_DataRegions) {
			dataView->view->setFont(font);
		}
	}
}

//------------------------------------------------------------------------------
// Name: finishPluginSetup(const QMap<QString, DebuggerPluginInterface *> &plugins)
// Desc: finalizes plugin setup by adding each to the menu, we can do this now 
//       that we have a GUI widget to attach it to
//------------------------------------------------------------------------------
void DebuggerMain::finishPluginSetup(const QMap<QString, DebuggerPluginInterface *> &plugins) {
	// setup the menu for all plugins that which to do so
	foreach (DebuggerPluginInterface *plugin, plugins) {
		QMenu *const menu = plugin->menu(this);
		if(menu != 0) {
			ui.menu_Plugins->addMenu(menu);
		}
	}
}

//------------------------------------------------------------------------------
// Name: setupTabButtons()
// Desc: creates the add/remove tab buttons in the data view
//------------------------------------------------------------------------------
void DebuggerMain::setupTabButtons() {
	// add the corner widgets to the data view
	m_AddTab = new QToolButton(ui.tabWidget);
	m_DelTab = new QToolButton(ui.tabWidget);
	
	m_AddTab->setToolButtonStyle(Qt::ToolButtonIconOnly);
	m_DelTab->setToolButtonStyle(Qt::ToolButtonIconOnly);
	
	m_AddTab->setIcon(QIcon(":/debugger/pics/addtab.png"));
	m_DelTab->setIcon(QIcon(":/debugger/pics/closetab.png"));
	m_AddTab->setAutoRaise(true);
	m_DelTab->setAutoRaise(true);
	m_DelTab->setEnabled(false);
	
	ui.tabWidget->setCornerWidget(m_AddTab, Qt::TopLeftCorner);
	ui.tabWidget->setCornerWidget(m_DelTab, Qt::TopRightCorner);
	
	connect(m_AddTab, SIGNAL(clicked()), SLOT(mnuDumpCreateTab()));
	connect(m_DelTab, SIGNAL(clicked()), SLOT(mnuDumpDeleteTab()));
}

//------------------------------------------------------------------------------
// Name: on_registerList_customContextMenuRequested(const QPoint &pos)
// Desc: context menu handler for register view
//------------------------------------------------------------------------------
void DebuggerMain::on_registerList_customContextMenuRequested(const QPoint &pos) {
	QTreeWidgetItem *const item = ui.registerList->itemAt(pos);
	if(item != 0) {
		if(!ui.registerList->isCatagory(item)) {
		
			// a little bit cheesy of a solution, but should work nicely
			edb::reg_t *const value = edb::v1::archProcessor().valueFromItem(*item, State::REG_GPR | State::REG_IP);
			if(value != 0) {		
				QMenu *menu = new QMenu(this);
				menu->addAction(tr("&Follow In Dump"), this, SLOT(mnuRegisterFollowInDump()));
				menu->addAction(tr("&Follow In Dump (New Tab)"), this, SLOT(mnuRegisterFollowInDumpNewTab()));
				menu->addAction(tr("&Follow In Stack"), this, SLOT(mnuRegisterFollowInStack()));
				menu->exec(ui.registerList->mapToGlobal(pos));
				release(menu);
			}
		}
	}
}

//------------------------------------------------------------------------------
// Name: on_cpuView_breakPointToggled(edb::address_t address)
// Desc: handler for toggling the breakpoints
//------------------------------------------------------------------------------
void DebuggerMain::on_cpuView_breakPointToggled(edb::address_t address) {
	edb::v1::toggleBreakpoint(address);
}

//------------------------------------------------------------------------------
// Name: on_registerList_itemDoubleClicked(QTreeWidgetItem *item)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_registerList_itemDoubleClicked(QTreeWidgetItem *item) {	
	Q_CHECK_PTR(item);
	edb::reg_t *const value = edb::v1::archProcessor().valueFromItem(*item, State::REG_ANY);
	
	if(value != 0) {
		if(edb::v1::getValueFromUser(*value, tr("Register Value"))) {			
			edb::v1::debuggerBase->setState(edb::v1::currentState());
			updateGUI();
		}
	}
}

//------------------------------------------------------------------------------
// Name: on_action_About_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_About_triggered() {
	QMessageBox::about(this, tr("About EDB"),
		QString(
			tr(
			"<p>EDB (Evan's Debugger) is designed to be an easy to use, modular, and cross platform debugger.</p>"
			"<p>More information and updates can be found at <a href=\"http://www.codef00.com\">http://www.codef00.com</a></p>"
			"<p>You can also report bugs an feature requests at <a href=\"http://bugs.codef00.com\">http://bugs.codef00.com</a></p>"
			"<p>Written by Evan Teran.</p>"
			"<p>version: %1</p>"
			)).arg(version_string));
}

//------------------------------------------------------------------------------
// Name: doApplyDefaultShowSeperator()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::doApplyDefaultShowSeperator() {
	const bool show = edb::v1::config().showAddressSeperator;
	
	ui.cpuView->setShowAddressSeparator(show);
	ui.stackView->setShowAddressSeparator(show);
	foreach(DataViewInfo *dataView, m_DataRegions) {
		dataView->view->setShowAddressSeparator(show);
	}
}

//------------------------------------------------------------------------------
// Name: on_action_Configure_Debugger_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Configure_Debugger_triggered() {
	
	if(edb::v1::dialogOptions()->exec() == QDialog::Accepted) {	

		// reload symbols in case they changed, or our symbol files changes
		edb::v1::reloadSymbols();
		
		// re-read the memory region information
		edb::v1::memoryRegions().sync();
		
		// apply the selected fonts
		doApplyDefaultFonts();
		
		// apply changes to the GUI options
		doApplyDefaultShowSeperator();
		
		// show changes
		refreshGUI();
	}
}

//------------------------------------------------------------------------------
// Name: getGotoExpression(bool &ok)
// Desc: 
//------------------------------------------------------------------------------
edb::address_t DebuggerMain::getGotoExpression(bool &ok) {
    const QString text = QInputDialog::getText(this, tr("Goto Address"), tr("Address:"), QLineEdit::Normal, QString(), &ok);
	
	if (ok && !text.isEmpty()) {
		Expression<edb::address_t> expr(text, &edb::v1::getVariable, 0, &edb::v1::getDWORD, 0);
		try {
			const edb::address_t address = expr.evaluateExpression();
			ok = true;
			return address;
		} catch(const Error &err) {
			QMessageBox::information(this, tr("Error In Address Expression!"), err.what());
			ok = false;
		}
	}
	return 0;
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::doMemoryFollow(bool ok, edb::address_t address, bool (*followFunc)(edb::address_t)) {
	if(ok && !(*followFunc)(address)) {
		QMessageBox::information(this, 
			tr("No Memory Found"), 
			QString(tr("There appears to be no memory at that location (<strong>0x%1</strong>)")).arg(address, EDB_MAX_HEX, 16, QChar('0')));
	}
}

//------------------------------------------------------------------------------
// Name: followRegisterInDump(bool tabbed)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::followRegisterInDump(bool tabbed) {
	bool ok;
	const edb::address_t address = getFollowRegister(ok);
	if(ok && !edb::v1::dumpData(address, tabbed)) {
		QMessageBox::information(this, 
			tr("No Memory Found"), 
			QString(tr("There appears to be no memory at that location (<strong>0x%1</strong>)")).arg(address, EDB_MAX_HEX, 16, QChar('0')));
	}
}

//------------------------------------------------------------------------------
// Name: mnuStackGotoESP()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuStackGotoESP() {
	doMemoryFollow(true, edb::v1::currentState().stackPointer(), edb::v1::dumpStack);
}

//------------------------------------------------------------------------------
// Name: mnuStackGotoEBP()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuStackGotoEBP() {
	doMemoryFollow(true, edb::v1::currentState().framePointer(), edb::v1::dumpStack);
}

//------------------------------------------------------------------------------
// Name: mnuCPUJumpToEIP()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuCPUJumpToEIP() {
	doMemoryFollow(true, edb::v1::currentState().instructionPointer(), edb::v1::jumpToAddress);
}

//------------------------------------------------------------------------------
// Name: mnuCPUJumpToAddress()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuCPUJumpToAddress() {
	bool ok;
	const edb::address_t address = getGotoExpression(ok);
	doMemoryFollow(ok, address, edb::v1::jumpToAddress);
}

//------------------------------------------------------------------------------
// Name: mnuDumpGotoAddress()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuDumpGotoAddress() {
    bool ok;
	const edb::address_t address = getGotoExpression(ok);
	doMemoryFollow(ok, address, edb::v1::dumpData);
}

//------------------------------------------------------------------------------
// Name: mnuStackGotoAddress()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuStackGotoAddress() {
    bool ok;
	const edb::address_t address = getGotoExpression(ok);
	doMemoryFollow(ok, address, edb::v1::dumpStack);
}

//------------------------------------------------------------------------------
// Name: mnuRegisterFollowInStack()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuRegisterFollowInStack() {
	bool ok;
	const edb::address_t address = getFollowRegister(ok);
	doMemoryFollow(ok, address, edb::v1::dumpStack);
}

//------------------------------------------------------------------------------
// Name: doFollowInStack(const QHexView *hv)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::doFollowInStack(const QHexView *hv) {
	bool ok;
	const edb::address_t address = getFollowAddress(hv, ok);
	doMemoryFollow(ok, address, edb::v1::dumpStack);
}

//------------------------------------------------------------------------------
// Name: doFollowInDump(const QHexView *hv)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::doFollowInDump(const QHexView *hv) {
	bool ok;
	const edb::address_t address = getFollowAddress(hv, ok);
	doMemoryFollow(ok, address, edb::v1::dumpData);
}

//------------------------------------------------------------------------------
// Name: getFollowRegister(bool &ok)
// Desc: 
//------------------------------------------------------------------------------
edb::reg_t DebuggerMain::getFollowRegister(bool &ok) {
	ok = false;
	const QTreeWidgetItem *const i = ui.registerList->currentItem();
	edb::reg_t ret = 0;
	if(i != 0) {
		edb::reg_t *const p = edb::v1::archProcessor().valueFromItem(*i, State::REG_GPR | State::REG_IP);

		if(p != 0) {
			ok = true;
			ret = *p;
		}
	}
	
	return ret;
}

//------------------------------------------------------------------------------
// Name: getFollowAddress(const QHexView *hv, bool &ok)
// Desc: 
//------------------------------------------------------------------------------
edb::address_t DebuggerMain::getFollowAddress(const QHexView *hv, bool &ok) {
	ok = false;
	
	Q_CHECK_PTR(hv);
	
	if(hv->hasSelectedText()) {
		const QByteArray data = hv->selectedBytes();

		if(data.size() == sizeof(edb::address_t)) {
			edb::address_t d;
			std::memcpy(&d, data.data(), sizeof(edb::address_t));

			ok = true;
			return d;
		}
	}
	
	QMessageBox::information(this, 
		tr("Invalid Selection"), 
		QString(tr("Please select a %1 bytes to use this function.")).arg(static_cast<unsigned>(sizeof(edb::address_t))));
		
	return 0;
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuDumpFollowInDump() {
	doFollowInDump(qobject_cast<QHexView *>(ui.tabWidget->currentWidget()));
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuDumpFollowInStack() {
	doFollowInStack(qobject_cast<QHexView *>(ui.tabWidget->currentWidget()));
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuStackFollowInDump() {
	doFollowInDump(ui.stackView);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuStackFollowInStack() {
	doFollowInStack(ui.stackView);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_actionApplication_Arguments_triggered() {
    bool ok;
    const QString text = QInputDialog::getText(this, tr("Application Arguments"), tr("Arguments:"), QLineEdit::Normal, m_Arguments, &ok);
	if(ok) {
		m_Arguments = text;
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_actionApplication_Working_Directory_triggered() {
    bool ok;
    const QString text = QInputDialog::getText(this, tr("Application Working Directory"), tr("Working Directory:"), QLineEdit::Normal, m_WorkingDirectory, &ok);
	if(ok) {
		m_WorkingDirectory = text;
	}
}

//------------------------------------------------------------------------------
// Name: mnuStackPush()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuStackPush() {

	// TODO: remove assuptions about stack growing downward

	edb::reg_t value;
	
	State &state = edb::v1::currentState();
	
	// get the value we are about to overwrite with the 'push'
	edb::v1::debuggerBase->readBytes(state.stackPointer() - sizeof(edb::reg_t), &value, sizeof(edb::reg_t));

	// ask for a replacement
	if(edb::v1::getValueFromUser(value, tr("Enter value to push"))) {

		// if they said ok, do the push, just like the hardware would do
		state.pushValue(value);

		// update the state
		edb::v1::debuggerBase->setState(state);
		updateGUI();
	}
}

//------------------------------------------------------------------------------
// Name: mnuStackPop()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuStackPop() {
	edb::v1::currentState().popValue();
	updateGUI();
}

//------------------------------------------------------------------------------
// Name: on_cpuView_customContextMenuRequested(const QPoint &pos)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_cpuView_customContextMenuRequested(const QPoint &pos) {
	QMenu *menu = new QMenu(this);
	// TODO: add comment
	//menu->addAction(tr("Add &Comment"), this, SLOT(""));
	//menu->addSeparator();
	menu->addAction(tr("&Goto Address"), this, SLOT(mnuCPUJumpToAddress()));
	menu->addAction(tr("&Goto EIP"), this, SLOT(mnuCPUJumpToEIP()));
	
	const edb::address_t address 	= ui.cpuView->selectedAddress();
	int size						= ui.cpuView->selectedSize();

	quint8 buffer[size];
	if(edb::v1::getInstructionBytes(address, buffer, size)) {
		edb::Instruction insn(buffer, size, address, std::nothrow);
		if(insn.valid()) {
		
			switch(insn.type()) {
			case edb::Instruction::OP_JMP:
			case edb::Instruction::OP_CALL:
			case edb::Instruction::OP_JCC:
				if(insn.operand(0).generalType() == edb::Operand::TYPE_REL) {
				
					QAction *const action = menu->addAction(tr("&Follow"), this, SLOT(mnuCPUFollow()));
					action->setData(static_cast<qlonglong>(insn.operand(0).relativeTarget()));
				}
				break;
			default:
				for(std::size_t i = 0; i < insn.operandCount(); ++i) {
					if(insn.operand(i).generalType() == edb::Operand::TYPE_IMMEDIATE) {
						
						QAction *const action = menu->addAction(tr("Follow Constant In &Dump"), this, SLOT(mnuCPUFollowInDump()));
						action->setData(static_cast<qlonglong>(insn.operand(i).immediate()));
						
						QAction *const action2 = menu->addAction(tr("Follow Constant In &Stack"), this, SLOT(mnuCPUFollowInStack()));
						action2->setData(static_cast<qlonglong>(insn.operand(i).immediate()));
					}
				}
			}
		}
	}
	
	menu->addSeparator();
	menu->addAction(tr("&Set EIP to this Instruction"), this, SLOT(mnuCPUSetEIP()));
	menu->addSeparator();
	menu->addAction(tr("&Edit Bytes"), this, SLOT(mnuCPUModify()));
	menu->addAction(tr("&Fill with 00's"), this, SLOT(mnuCPUFillZero()));
	menu->addAction(tr("&Fill with NOPs"), this, SLOT(mnuCPUFillNop()));
	menu->addSeparator();
	menu->addAction(tr("&Add Breakpoint"), this, SLOT(mnuCPUAddBreakpoint()));
	menu->addAction(tr("&Add Conditional Breakpoint"), this, SLOT(mnuCPUAddConditionalBreakpoint()));
	menu->addAction(tr("&Remove Breakpoint"), this, SLOT(mnuCPURemoveBreakpoint()));
	
	menu->exec(ui.cpuView->viewport()->mapToGlobal(pos));
	release(menu);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuCPUFollow() {
	QAction *const action = qobject_cast<QAction *>(sender());
	if(action != 0) {
		const edb::address_t address = action->data().toULongLong();
		doMemoryFollow(true, address, edb::v1::jumpToAddress);
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuCPUFollowInDump() {
	QAction *const action = qobject_cast<QAction *>(sender());
	if(action != 0) {
		const edb::address_t address = action->data().toULongLong();
		doMemoryFollow(true, address, edb::v1::dumpData);
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuCPUFollowInStack() {
	QAction *const action = qobject_cast<QAction *>(sender());
	
	if(action != 0) {
		const edb::address_t address = action->data().toULongLong();
		doMemoryFollow(true, address, edb::v1::dumpStack);
	}
}


//------------------------------------------------------------------------------
// Name: mnuStackToggleLock(bool locked)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuStackToggleLock(bool locked) {
	m_StackViewLocked = locked;
}

//------------------------------------------------------------------------------
// Name: on_stackView_customContextMenuRequested(const QPoint &pos)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_stackView_customContextMenuRequested(const QPoint &pos) {
	QHexView *const s = ui.stackView;
	Q_CHECK_PTR(s);
	
	QMenu *menu = s->createStandardContextMenu();
	menu->addSeparator();
	menu->addAction(tr("Follow Address In &Dump"), this, SLOT(mnuStackFollowInDump()));
	menu->addAction(tr("Follow Address In &Stack"), this, SLOT(mnuStackFollowInStack()));
	menu->addAction(tr("&Goto Address"), this, SLOT(mnuStackGotoAddress()));
	menu->addAction(tr("Goto E&SP"), this, SLOT(mnuStackGotoESP()));
	menu->addAction(tr("Goto E&BP"), this, SLOT(mnuStackGotoEBP()));
	menu->addSeparator();
	menu->addAction(tr("&Edit Bytes"), this, SLOT(mnuStackModify()));
	menu->addSeparator();
	menu->addAction(tr("&Push DWORD"), this, SLOT(mnuStackPush()));
	menu->addAction(tr("P&op DWORD"), this, SLOT(mnuStackPop()));
	
	// lockable stack feature
	menu->addSeparator();
	QAction *const action = new QAction(tr("&Lock Stack"), menu);
    action->setCheckable(true);
    action->setChecked(m_StackViewLocked);
	menu->addAction(action);
	connect(action, SIGNAL(toggled(bool)), SLOT(mnuStackToggleLock(bool)));
	
	menu->exec(s->mapToGlobal(pos));
	release(menu);
}

//------------------------------------------------------------------------------
// Name: mnuDumpContextMenu(const QPoint &pos)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuDumpContextMenu(const QPoint &pos) {
	QHexView *const s = qobject_cast<QHexView *>(sender());
	
	Q_CHECK_PTR(s);
	
	QMenu *menu = s->createStandardContextMenu();
	menu->addSeparator();
	menu->addAction(tr("Follow Address In &Dump"), this, SLOT(mnuDumpFollowInDump()));
	menu->addAction(tr("Follow Address In &Stack"), this, SLOT(mnuDumpFollowInStack()));
	menu->addAction(tr("&Goto Address"), this, SLOT(mnuDumpGotoAddress()));
	menu->addSeparator();
	menu->addAction(tr("&Edit Bytes"), this, SLOT(mnuDumpModify()));
	menu->addSeparator();
	menu->addAction(tr("&Save To File"), this, SLOT(mnuDumpSaveToFile()));
	menu->exec(s->mapToGlobal(pos));
	release(menu);
}

//------------------------------------------------------------------------------
// Name: mnuDumpSaveToFile(
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuDumpSaveToFile() {
	QHexView *const s = qobject_cast<QHexView *>(ui.tabWidget->currentWidget());
	
	Q_CHECK_PTR(s);

	const QString filename = QFileDialog::getSaveFileName(this, tr("Save File"), m_LastOpenDirectory, QString());
	if(!filename.isEmpty()) {

		QFile file(filename);
		file.open(QIODevice::WriteOnly);
		if(file.isOpen()) {
			file.write(s->allBytes());
		}			
	}
}

//------------------------------------------------------------------------------
// Name: overwriteCheck(edb::address_t address, unsigned int size)
// Desc: 
//------------------------------------------------------------------------------
bool DebuggerMain::overwriteCheck(edb::address_t address, unsigned int size) {
	bool firstConflict = true;
	for(edb::address_t addr = address; addr != (address + size); ++addr) {
		Breakpoint *const bp = edb::v1::debuggerBase->findBreakpoint(addr);

		if(bp != 0 && bp->enabled()) {
			if(firstConflict) {
			
				const int ret = QMessageBox::question(
						this, 
						tr("Overwritting breakpoint"), 
						tr("You are attempting to modify bytes which overlap with a software breakpoint. Doing this will implicitly remove any breakpoints which are a conflict. Are you sure you want to do this?"), 
						QMessageBox::Yes, 
						QMessageBox::No, 
						QMessageBox::NoButton);
						
				if(ret == QMessageBox::No) {
					return false;
				}
				firstConflict = false;
			}

			edb::v1::removeBreakpoint(addr);
		}
	}
	return true;
}

//------------------------------------------------------------------------------
// Name: doModifyBytesBase(edb::address_t address, unsigned int size, QByteArray &bytes, quint8 fill
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::doModifyBytesBase(edb::address_t address, unsigned int size, QByteArray &bytes, quint8 fill) {

	if(size != 0) {
		if(edb::v1::getBinaryStringFromUser(bytes, tr("Edit Binary String"), size)) {			
			if(overwriteCheck(address, size)) {

				// fill bytes
				while(bytes.size() < static_cast<int>(size)) {
					bytes.push_back(fill);
				}

				edb::v1::debuggerBase->writeBytes(address, bytes.data(), size);

				// do a refresh, not full update
				refreshGUI();
			}
		}
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::doCPUFill(quint8 byte) {
	const edb::address_t address 	= ui.cpuView->selectedAddress();
	const unsigned int size			= ui.cpuView->selectedSize();
	
	if(size != 0) {
		if(overwriteCheck(address, size)) {
			QByteArray bytes(size, byte);

			edb::v1::debuggerBase->writeBytes(address, bytes.data(), size);

			// do a refresh, not full update
			refreshGUI();
		}
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuCPUAddBreakpoint() {
	const edb::address_t address = ui.cpuView->selectedAddress();
	edb::v1::createBreakpoint(address);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuCPUAddConditionalBreakpoint() {
	bool ok;
	const edb::address_t address = ui.cpuView->selectedAddress();	
	const QString condition = QInputDialog::getText(this, tr("Set Breakpoint Condition"), tr("Expression:"), QLineEdit::Normal, QString(), &ok);
	if(ok) {
		edb::v1::createBreakpoint(address);
		if(!condition.isEmpty()) {
			edb::v1::debuggerBase->setBreakpointCondition(address, condition);
		}
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuCPURemoveBreakpoint() {
	const edb::address_t address = ui.cpuView->selectedAddress();
	edb::v1::removeBreakpoint(address);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuCPUFillZero() {
	doCPUFill(0x00);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuCPUFillNop() {
	doCPUFill(0x90);
}

//------------------------------------------------------------------------------
// Name: mnuCPUSetEIP()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuCPUSetEIP() {
	const edb::address_t address = ui.cpuView->selectedAddress();
	State &state = edb::v1::currentState();
	state.setInstructionPointer(address);
	edb::v1::debuggerBase->setState(state);
	updateGUI();
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuCPUModify() {
	const edb::address_t address 	= ui.cpuView->selectedAddress();
	const unsigned int size			= ui.cpuView->selectedSize();
	
	quint8 buf[edb::Instruction::MAX_SIZE];		
	
	Q_ASSERT(size <= sizeof(buf));
	
	const bool ok = edb::v1::debuggerBase->readBytes(address, buf, size);
	if(ok) {
		QByteArray bytes = QByteArray::fromRawData(reinterpret_cast<const char *>(buf), size);	
		doModifyBytesBase(address, size, bytes, 0x00);
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::doModifyBytes(QHexView *s) {
	if(s != 0) {
		const edb::address_t address	= s->selectedBytesAddress();
		const unsigned int size 		= s->selectedBytesSize();
		QByteArray bytes				= s->selectedBytes();
		
		doModifyBytesBase(address, size, bytes, 0x00);
	}
}

//------------------------------------------------------------------------------
// Name: mnuDumpModify()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuDumpModify() {
	doModifyBytes(qobject_cast<QHexView *>(ui.tabWidget->currentWidget()));
}

//------------------------------------------------------------------------------
// Name: mnuStackModify()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuStackModify() {
	doModifyBytes(ui.stackView);
}

//------------------------------------------------------------------------------
// Name: updateMenuState(GUI_STATE guiState)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::updateMenuState(GUI_STATE guiState) {

	switch(guiState) {
	case PAUSED:
	#if 0
		ui.actionRun_Until_Return->setEnabled(true);
	#endif
		ui.action_Restart->setEnabled(true);
		ui.action_Run->setEnabled(true);
		ui.action_Pause->setEnabled(false);
		ui.action_Step_Into->setEnabled(true);
		ui.action_Step_Over->setEnabled(true);
		ui.action_Step_Into_Pass_Signal_To_Application->setEnabled(true);
		ui.action_Step_Over_Pass_Signal_To_Application->setEnabled(true);
		ui.action_Run_Pass_Signal_To_Application->setEnabled(true);
		ui.action_Detach->setEnabled(true);
		ui.action_Kill->setEnabled(true);
		break;
	case RUNNING:
	#if 0
		ui.actionRun_Until_Return->setEnabled(false);
	#endif
		ui.action_Restart->setEnabled(false);
		ui.action_Run->setEnabled(false);
		ui.action_Pause->setEnabled(true);
		ui.action_Step_Into->setEnabled(false);
		ui.action_Step_Over->setEnabled(false);
		ui.action_Step_Into_Pass_Signal_To_Application->setEnabled(false);
		ui.action_Step_Over_Pass_Signal_To_Application->setEnabled(false);
		ui.action_Run_Pass_Signal_To_Application->setEnabled(false);
		ui.action_Detach->setEnabled(true);
		ui.action_Kill->setEnabled(true);
		break;
	case TERMINATED:
	#if 0
		ui.actionRun_Until_Return->setEnabled(false);
	#endif
		ui.action_Restart->setEnabled(false);
		ui.action_Run->setEnabled(false);
		ui.action_Pause->setEnabled(false);
		ui.action_Step_Into->setEnabled(false);
		ui.action_Step_Over->setEnabled(false);
		ui.action_Step_Into_Pass_Signal_To_Application->setEnabled(false);
		ui.action_Step_Over_Pass_Signal_To_Application->setEnabled(false);
		ui.action_Run_Pass_Signal_To_Application->setEnabled(false);
		ui.action_Detach->setEnabled(false);
		ui.action_Kill->setEnabled(false);
		break;
	}
	
	m_GUIState = guiState;
}

//------------------------------------------------------------------------------
// Name: handleEventSignaled()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::handleEventSignaled() {
	switch(m_LastEvent.exitSignal()) {
	case DebugEvent::sigkill:
		QMessageBox::information(this, tr("Application Killed"), tr("The debugged application was killed."));
		break;
	default:
		qDebug() << "Unknown Termination Signal Recieved! " << m_LastEvent.exitSignal();
		break;
	}

	on_action_Detach_triggered();
}

//------------------------------------------------------------------------------
// Name: doTrap()
// Desc: returns true if we should resume as if this trap never happened
//------------------------------------------------------------------------------
bool DebuggerMain::doTrap() {

	// we just got a DebugEvent::sigtrap, there are a few possible causes
	// #1: we just did an execve, in which case we want to stop cause we just attached
	// #2: we hit a 0xCC breakpoint, if so, then we want to stop
	// #3: we did a step
	// #4: we are tracing
	// #5: we hit a hardware breakpoint!
	
	State &state = edb::v1::currentState();

	// prevEIP will have the address of the instruction which brought us here!
	const edb::address_t prevEIP = state.instructionPointer() - edb::v1::debuggerBase->executionBreakpointSize();

	// look it up in our breakpoint list, make sure it is one of OUR int3s!
	// if it is, we need to backup EIP and pause ourselves
	Breakpoint *const bp = edb::v1::debuggerBase->findBreakpoint(prevEIP);
	if(bp != 0 && bp->enabled()) {
	
		// TODO: check if the breakpoint was corrupted

		++bp->passcount;					

		// back up eip the size of a breakpoint, since we executed a breakpoint 
		// instead of the real code that belongs there
		state.setInstructionPointer(prevEIP);
		edb::v1::debuggerBase->setState(state);
		
		const QString bpCondition = bp->condition;

		// handle conditional breakpoints
		if(m_StepWasRun && !bpCondition.isEmpty()) {

			int conditionVal = 0;
			Expression<edb::address_t> expr(bpCondition, &edb::v1::getVariable, 0, &edb::v1::getDWORD, 0);
			try {
				conditionVal = expr.evaluateExpression();
			} catch(const Error &err) {
				QMessageBox::information(this, tr("Error In Breakpoint Expression!"), err.what());
				conditionVal = 1;
				// TODO: remove the breakpoint
			}				

			// ok, the condition was false, so skip over this breakpoint...
			if(!conditionVal) {
				return true;
			}
		}
		
		// if it's a one time breakpoint then we should remove it upon
		// triggering, this is mainly used for 'run-until' situations
		// like step over

		if(bp->oneTime()) {
			edb::v1::debuggerBase->removeBreakpoint(prevEIP);
		}
		
		// cancel any runs we were going to do..
		m_StepWasRun = false;
	}
	
	return false;
}

//------------------------------------------------------------------------------
// Name: handleEventStopped()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::handleEventStopped() {


	// ok we just came in from a signal, we need to test some things, 
	// generally, we will want to check if it was a step, if it was, was it 
	// because we just hit a break point or because we wanted to run, but had 
	// to step first in case were were on a breakpoint already...

	switch(m_LastEvent.stopSignal()) {
	case DebugEvent::sigsegv:
		QMessageBox::information(this, tr("Illegal Access Fault"), 
			QString(tr(
				"<p>The debugged application encountered a segmentation fault!<br />"
				"The address <strong>0x%1</strong> could not be accessed.</p>"
				"<p>If you would like to pass this exception to the application press Shift+[F7/F8/F9]</p>")
				).arg(edb::v1::debuggerBase->getFaultAddress(), EDB_MAX_HEX, 16, QChar('0')));
		m_StepWasRun = false;
		break;
		
	case DebugEvent::sigill:
		QMessageBox::information(this, tr("Illegal Instruction Fault"),
			tr("<p>The debugged application attempted to execute an illegal instruction.</p>"
			"<p>If you would like to pass this exception to the application press Shift+[F7/F8/F9]</p>")
			);
		m_StepWasRun = false;
		break;
		
	case DebugEvent::sigfpe:
		QMessageBox::information(this, tr("Floating Point Exception"), 
			tr("<p>The debugged application encountered a floating-point exception!</p>"
			"<p>If you would like to pass this exception to the application press Shift+[F7/F8/F9]</p>")
			);
		m_StepWasRun = false;
		break;
		
	case DebugEvent::sigabrt:
		QMessageBox::information(this, tr("Abort"), 
			tr("<p>The debugged application has aborted!</p>"
			"<p>If you would like to pass this exception to the application press Shift+[F7/F8/F9]</p>")
			);
		m_StepWasRun = false;
		break;
		
	case DebugEvent::sigstop:
		// user asked to pause the debugged process
		m_StepWasRun = false;
		break;
		
	case DebugEvent::sigtrap:
		
		if(doTrap()) {
			doResume(0);
			return;
		}
		break;

	case DebugEvent::sigchld:
		// i doubt we should see these
		edb::v1::debuggerBase->resume(m_LastEvent.stopSignal());
		return;
		
	case DebugEvent::sigcont:
		// this one is probably not very interesting, so we'll just resume
		edb::v1::debuggerBase->resume(m_LastEvent.stopSignal());
		return;

	case DebugEvent::sigalrm:
		qDebug("ALRM Stop Signal Recieved!");
		if(m_StepWasRun) {
			edb::v1::debuggerBase->resume(m_LastEvent.stopSignal());
			return;
		}
		break;
			
	default:
		qDebug() << "Unknown Stop Signal Recieved!" << m_LastEvent.stopSignal();
		if(m_StepWasRun) {
			edb::v1::debuggerBase->resume(m_LastEvent.stopSignal());
			return;
		}
		break;
		
	}
	
	edb::v1::debuggerBase->enableBreakpoint(m_ReenableBreakpointAddress);
	m_ReenableBreakpointAddress = 0;

	// if we stepped with the intention of running, well we should run then!
	if(m_StepWasRun) {
		m_StepWasRun = false;
		edb::v1::debuggerBase->resume(0);
	} else {
		updateGUI();
		updateMenuState(PAUSED);
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::handleEventExited() {
	QMessageBox::information(this, tr("Application Exited"), QString(tr("The debugged application exited normally with exit code %1.")).arg(m_LastEvent.exitStatus()));
	on_action_Detach_triggered();
}

//------------------------------------------------------------------------------
// Name: handleEvent(DebugEvent event)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::handleEvent(DebugEvent event) {
	
	Q_CHECK_PTR(edb::v1::debuggerBase);
	
	edb::v1::debuggerBase->getState(edb::v1::currentState());
	edb::v1::memoryRegions().sync();
	
	m_LastEvent = event;
	
	switch(m_LastEvent.reason()) {
	case DebugEvent::EVENT_STOPPED:
		handleEventStopped();
		break;
		
	case DebugEvent::EVENT_SIGNALED:
		handleEventSignaled();
		break;
		
	case DebugEvent::EVENT_EXITED:
		handleEventExited();
		break;
		
	case DebugEvent::EVENT_UNKNOWN:
		break;
	}
}

//------------------------------------------------------------------------------
// Name: debugEventHandler(DebugEvent event)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::debugEventHandler(DebugEvent event) {
	DebugEventHandlerInterface *const handler = edb::v1::debugEventHandler();
	Q_CHECK_PTR(handler);
	handler->handleEvent(event);
}

//------------------------------------------------------------------------------
// Name: update()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::DataViewInfo::update() {

	Q_CHECK_PTR(stream);
	Q_CHECK_PTR(view);
	
	stream->setRegion(region);	
	view->setAddressOffset(region.start);
	view->setData(stream);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::updateTabCaption(QHexView *view, edb::address_t start, edb::address_t end) {
	const int index = ui.tabWidget->indexOf(view);
	const QString caption = qobject_cast<ETabWidget *>(ui.tabWidget)->caption(index);
	
	if(caption.isEmpty()) {
		ui.tabWidget->setTabText(index, QString("%1-%2").arg(start, EDB_MAX_HEX, 16, QChar('0')).arg(end, EDB_MAX_HEX, 16, QChar('0')));
	} else {
		ui.tabWidget->setTabText(index, QString("[%1] %2-%3").arg(caption).arg(start, EDB_MAX_HEX, 16, QChar('0')).arg(end, EDB_MAX_HEX, 16, QChar('0')));
	}
}

//------------------------------------------------------------------------------
// Name: updateData(DebuggerMain::DataViewInfo *v)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::updateData(DebuggerMain::DataViewInfo *v) {

	Q_CHECK_PTR(v);
	QHexView *const view = v->view;
	
	Q_CHECK_PTR(view);

	v->update();
	
	updateTabCaption(view, v->region.start, v->region.end);
}

//------------------------------------------------------------------------------
// Name: clearData(DataViewInfo *v)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::clearData(DataViewInfo *v) {

	Q_CHECK_PTR(v);
	
	QHexView *const view = v->view;
	
	Q_CHECK_PTR(view);

	view->clear();
	view->scrollTo(0);
	
	updateTabCaption(view, edb::nullptr, edb::nullptr);
}

//------------------------------------------------------------------------------
// Name: doJumpToAddress(edb::address_t address, const MemoryRegions::Region &r, bool scrollTo)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::doJumpToAddress(edb::address_t address, const MemoryRegions::Region &r, bool scrollTo) {

	ui.cpuView->setAddressOffset(r.start);
	ui.cpuView->setRegion(r);
	if(scrollTo && !ui.cpuView->addressShown(address)) {
		ui.cpuView->scrollTo(address);
	}
}

//------------------------------------------------------------------------------
// Name: updateDisassembly(edb::address_t address, const MemoryRegions::Region &r)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::updateDisassembly(edb::address_t address, const MemoryRegions::Region &r) {

	ui.cpuView->setCurrentAddress(address);	
	doJumpToAddress(address, r, true);
	m_ListModel->setStringList(edb::v1::archProcessor().updateInstructionInfo(address));
}

//------------------------------------------------------------------------------
// Name: updateStackView()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::updateStackView() {
	if(!edb::v1::dumpStack(edb::v1::currentState().stackPointer(), !m_StackViewLocked)) {
		ui.stackView->clear();
		ui.stackView->scrollTo(0);
	}
}

//------------------------------------------------------------------------------
// Name: updateCPUView(MemoryRegions::Region &region)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::updateCPUView(MemoryRegions::Region &region) {

	const edb::address_t address = edb::v1::currentState().instructionPointer();

	if(!edb::v1::memoryRegions().findRegion(address, region)) {
		ui.cpuView->clear();
		ui.cpuView->scrollTo(0);
		m_ListModel->setStringList(QStringList());
	} else {
		updateDisassembly(address, region);
	}
}

//------------------------------------------------------------------------------
// Name: updateDataViews()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::updateDataViews() {

	// update all data views with the current region data
	foreach(DataViewInfo *info, m_DataRegions) {

		// make sure the regions are still valid..
		if(edb::v1::memoryRegions().findRegion(info->region.start)) {
			updateData(info);
		} else {
			clearData(info);
		}
	}
}

//------------------------------------------------------------------------------
// Name: refreshGUI()
// Desc: refreshed all the different displays
//------------------------------------------------------------------------------
void DebuggerMain::refreshGUI() {
	ui.cpuView->repaint();
	ui.stackView->repaint();
	
	foreach(DataViewInfo *info, m_DataRegions) {
		QHexView *const view = info->view;
	
		Q_CHECK_PTR(view);
		view->repaint();
	}
	
	m_ListModel->setStringList(edb::v1::archProcessor().updateInstructionInfo(edb::v1::currentState().instructionPointer()));
}

//------------------------------------------------------------------------------
// Name: updateGUI()
// Desc: updates all the different displays
//------------------------------------------------------------------------------
void DebuggerMain::updateGUI() {

	MemoryRegions::Region region;
	updateCPUView(region);
	updateDataViews();
	updateStackView();
	edb::v1::archProcessor().updateRegisterView(region.name);
}

//------------------------------------------------------------------------------
// Name: doResume(bool passExceptionToApp)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::doResume(bool passExceptionToApp) {
	
	// if we are on a breakpoint, disable it, and set the re-enable 
	// address to the current eip
	// disableBreakpoint returns 0 if there was no breakpoint to disable
	m_ReenableBreakpointAddress = edb::v1::debuggerBase->disableBreakpoint(edb::v1::currentState().instructionPointer());

	// set the state to 'running'
	updateMenuState(RUNNING);
	
	// if neccessary pass the signal to the application, otherwise just step
	if(passExceptionToApp && m_LastEvent.stopped() && m_LastEvent.stopSignal() != DebugEvent::sigtrap && m_LastEvent.stopSignal() != DebugEvent::sigstop) {
		edb::v1::debuggerBase->step(m_LastEvent.stopSignal());
	} else {
		edb::v1::debuggerBase->step(0);
	}
}

//------------------------------------------------------------------------------
// Name: on_action_Run_Pass_Signal_To_Application_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Run_Pass_Signal_To_Application_triggered() {
	m_StepWasRun = true;
	doResume(true);	
}

//------------------------------------------------------------------------------
// Name: on_action_Step_Into_Pass_Signal_To_Application_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Step_Into_Pass_Signal_To_Application_triggered() {
	doResume(true);
}

//------------------------------------------------------------------------------
// Name: on_action_Step_Over_Pass_Signal_To_Application_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Step_Over_Pass_Signal_To_Application_triggered() {
	doStepOver<
		&DebuggerMain::on_action_Run_Pass_Signal_To_Application_triggered, 
		&DebuggerMain::on_action_Step_Into_Pass_Signal_To_Application_triggered>();
}

//------------------------------------------------------------------------------
// Name: on_action_Step_Over_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Step_Over_triggered() {
	doStepOver<
		&DebuggerMain::on_action_Run_triggered, 
		&DebuggerMain::on_action_Step_Into_triggered>();
}

//------------------------------------------------------------------------------
// Name: on_action_Run_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Run_triggered() {
	m_StepWasRun = true;
	doResume(false);
}

//------------------------------------------------------------------------------
// Name: on_action_Step_Into_triggered(0
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Step_Into_triggered() {
	doResume(false);
}

//------------------------------------------------------------------------------
// Name: on_actionRun_Until_Return_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_actionRun_Until_Return_triggered() {
	// TODO: implement this
	// TODO: we need a generic 'run until <condition>' function
	// likely callback based
}

//------------------------------------------------------------------------------
// Name: on_action_Pause_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Pause_triggered() {
	Q_CHECK_PTR(edb::v1::debuggerBase);
	edb::v1::debuggerBase->pause();
}

//------------------------------------------------------------------------------
// Name: cleanupEventThread()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::cleanupEventThread() {

	if(m_DebugEventServer != 0) {
	
		Q_CHECK_PTR(edb::v1::debuggerBase);
	
		// make sure we don't get any spurious events!
		m_DebugEventServer->disconnect();
			
		// make sure we are paused...
		if(m_GUIState == RUNNING) {
			edb::v1::debuggerBase->pause();
			updateMenuState(PAUSED);
		}
		
		// remove all breakpoints, we need to do this during this moment so that 
		// we don't get any signals in the child process later!
		edb::v1::debuggerBase->clearBreakpoints();
		
		// make sure the server thread thinks it should exit
		m_DebugEventServer->m_ShouldExit = true;

		// this is ugly, but it seems to work pretty well
		// basically we have marked the thread as 'should exit'
		// but, we may be blocked (likley) waiting for the next event
		// so we intentionally cause one last event giving it a chance to
		// check the loop invariant one more time, which will allow
		// the thread to exit cleanly
		if(m_GUIState == PAUSED) {
			edb::v1::debuggerBase->step(0);
			updateMenuState(RUNNING);
		}
				
		m_DebugEventServer->wait();
		m_DebugEventServer = 0;
	}
}

//------------------------------------------------------------------------------
// Name: cleanupDebugger()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::cleanupDebugger() {
	
	cleanupEventThread();

	edb::v1::memoryRegions().clear();
	edb::v1::symbolManager().clear();
	edb::v1::currentState().clear();
	edb::v1::archProcessor().reset();

	// clear up the data view
	while(ui.tabWidget->count() > 1) {
		mnuDumpDeleteTab();
	}
	
	Q_ASSERT(!m_DataRegions.isEmpty());
	m_DataRegions.first()->region = MemoryRegions::Region();
		
	setWindowTitle(tr("EDB"));
	
	updateGUI();
}

//------------------------------------------------------------------------------
// Name: sessionFileName() const
// Desc: 
//------------------------------------------------------------------------------
QString DebuggerMain::sessionFileName() const {
	QString ret;
	if(!m_ProgramExecutable.isEmpty()) {
		const QString basename = edb::v1::basename(m_ProgramExecutable);
		ret = edb::v1::config().sessionPath + "/" + basename + ".edb_session";
	}
	return ret;
}

//------------------------------------------------------------------------------
// Name: doDetachFromProcess(bool kill)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::doDetachFromProcess(bool kill) {
	
	if(edb::v1::debuggerBase != 0) {
	
	#ifdef USE_SESSION
		const QString sessionFile = sessionFileName();
		if(!sessionFile.isEmpty()) {
			m_SessionFileManager.saveSessionData(sessionFile, m_ProgramExecutable);
			m_ProgramExecutable.clear();
		}
	#endif
		cleanupDebugger();

		if(kill) {
			edb::v1::debuggerBase->kill();
		} else {
			edb::v1::debuggerBase->detach();
		}
	}

	updateMenuState(TERMINATED);
}

//------------------------------------------------------------------------------
// Name: on_action_Detach_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Detach_triggered() {
	doDetachFromProcess(false);
}

//------------------------------------------------------------------------------
// Name: on_action_Kill_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Kill_triggered() {
	doDetachFromProcess(true);
}

//------------------------------------------------------------------------------
// Name: setInitialDebuggerState()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::setInitialDebuggerState() {

	updateMenuState(PAUSED);
		
	m_DebugEventServer = new DebugEventServer(this);
	connect(m_DebugEventServer, SIGNAL(debugEvent(DebugEvent)), SLOT(debugEventHandler(DebugEvent)));
	m_DebugEventServer->start();
	
	edb::v1::symbolManager().loadSymbols(edb::v1::config().symbolPath);
	
	edb::v1::memoryRegions().setPID(edb::v1::debuggerBase->pid());
	edb::v1::memoryRegions().sync();
	
	const QVector<MemoryRegions::Region> &l = edb::v1::memoryRegions().regions();

	Q_ASSERT(m_DataRegions.size() > 0);

	if(l.size() >= 2) {
		m_DataRegions.first()->region = l[1];
	} else {
		m_DataRegions.first()->region = MemoryRegions::Region();
	}
	
	edb::v1::currentState().clear();
	
	m_StepWasRun = false;
	
	setDebuggerCaption(edb::v1::getProcessExecutable());
	
#ifdef USE_SESSION
	
	m_ProgramExecutable.clear();
	if(edb::v1::debuggerBase->pid() != 0) {
		m_ProgramExecutable	= edb::v1::getProcessExecutable();		
	}

	const QString sessionFile = sessionFileName();
	if(!sessionFile.isEmpty()) {
		m_SessionFileManager.loadSessionData(sessionFile);
	}
#endif
}

//------------------------------------------------------------------------------
// Name: setInitialBreakpoint(const QString &s)
// Desc: sets the initial breakpoint so we can stop at the entry point of the 
//       application
//------------------------------------------------------------------------------
void DebuggerMain::setInitialBreakpoint(const QString &s) {
	edb::address_t entryPoint = 0;
	
	if(edb::v1::config().initialBreakpoint == Configuration::MainSymbol) {
		const QString mainSymbol = edb::v1::basename(s) + "::main";
		const SymbolManager::Symbol *const sym = edb::v1::symbolManager().findSymbol(mainSymbol);

		if(sym != 0) {
			entryPoint = sym->address;
		} else if(edb::v1::config().findMain) {
			// try to figure out where main would be
			entryPoint = edb::v1::locateMainFunction(s);
		}

	} else if(edb::v1::config().initialBreakpoint == Configuration::EntryPoint) {

		BinaryInfo *binfo = edb::v1::getBinaryInfo(s);
		if(binfo != 0) {
			entryPoint = binfo->getEntryPoint();
		}
		release(binfo);
	}

	if(entryPoint != 0) {
		edb::v1::debuggerBase->addBreakpoint(entryPoint, true);
	}
}

//------------------------------------------------------------------------------
// Name: ttyProcFinished(int exitCode, QProcess::ExitStatus exitStatus)
// Desc: cleans up the data associated with a TTY when the terminal dies
//------------------------------------------------------------------------------
void DebuggerMain::ttyProcFinished(int exitCode, QProcess::ExitStatus exitStatus) {
	Q_UNUSED(exitCode);
	Q_UNUSED(exitStatus);
	
	m_TTYFile.clear();
	release(m_TTYProc);
}

//------------------------------------------------------------------------------
// Name: createTTY()
// Desc: creates a TTY object for our command line I/O
//------------------------------------------------------------------------------
QString DebuggerMain::createTTY() {
#ifdef Q_OS_LINUX
	// we attempt to reuse an open output window
	if(m_TTYProc == 0 && edb::v1::config().ttyEnabled) {
	
		// ok, creating a new one...	
		// first try to get a 'unique' filename, i would love to use a system 
		// temp file API... but there doesn't seem to be one which will create 
		// a pipe...only ordinary files!
		const QString tempPipe = QString("%1/edb_temp_file_%2_%3").arg(QDir::tempPath()).arg(rand()).arg(getpid());

		// make sure it isn't already there, and then make the pipe
		::unlink(qPrintable(tempPipe));
		::mkfifo(qPrintable(tempPipe), S_IRUSR | S_IWUSR);

		// this is a basic shell script which will output the tty to a file (the pipe),
		// ignore kill sigs, close all standard IO, and then just hang
		const QString shellScript = QString(
		    "tty > %1;"
			"trap \"\" INT QUIT TSTP;"
			"exec<&-; exec>&-;"
			"while :; do sleep 3600; done"
			).arg(tempPipe);

		// parse up the command from the options, white space delimited
		QStringList procargs = edb::v1::parseCommandLine(edb::v1::config().ttyCommand);
		const QString ttyCommand = procargs.takeFirst();

		// start constructing the arguments for the term
		procargs << "-title" <<  tr("EDB Output") << "-e" << "sh" <<  "-c" << shellScript;

		// make the tty process object and connect it's death signal to our cleanup
		m_TTYProc = new QProcess(this);
		connect(m_TTYProc, SIGNAL(finished(int, QProcess::ExitStatus)), SLOT(ttyProcFinished(int, QProcess::ExitStatus)));

		m_TTYProc->start(ttyCommand, procargs);
		const bool ok = m_TTYProc->waitForStarted();
				
		if(ok) {
			// get the output, this should block until the xterm actually gets a chance to write it
			QFile file(tempPipe);
			if(file.open(QIODevice::ReadOnly)) {
				m_TTYFile = file.readLine().trimmed();
			}
		} else {
			qDebug().nospace() << "Could not launch the desired terminal [" << ttyCommand << "], please check that it exists and you have proper permissions.";
			release(m_TTYProc);
			m_TTYFile.clear();
		}

		// cleanup, god i wish there was an easier way than this!
		::unlink(qPrintable(tempPipe));
	}
#endif	
	return m_TTYFile;
}

//------------------------------------------------------------------------------
// Name: on_action_Restart_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Restart_triggered() {
	
	Q_CHECK_PTR(edb::v1::debuggerBase);

	m_WorkingDirectory = edb::v1::getProcessCurrentWorkingDirectory();	
	QStringList args = edb::v1::getProcessArguments();

	// args must have at least one element because the first is the program
	// name
	if(args.size() != 0) {

		const QString s = edb::v1::getProcessExecutable();
		
		args.removeFirst();

		if(!s.isEmpty()) {
			doDetachFromProcess(true);
			commonOpen(s, args);
		}
	}
}

//------------------------------------------------------------------------------
// Name: commonOpen(const QString &s, const QStringList &args)
// Desc: 
//------------------------------------------------------------------------------
bool DebuggerMain::commonOpen(const QString &s, const QStringList &args) {
	const QString tty = createTTY();
	bool ret = false;
	if(edb::v1::debuggerBase->open(s, m_WorkingDirectory, args, tty)) {
		setInitialDebuggerState();
		setInitialBreakpoint(s);
		ret = true;		
	} else {
		QMessageBox::information(this, tr("Could Not Open"), tr("Failed to open and attach to process, please check priviledges and try again."));
	}

	edb::v1::debuggerBase->getState(edb::v1::currentState());
	updateGUI();
	return ret;
}

//------------------------------------------------------------------------------
// Name: doRun(const QString &s, const QStringList &args)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::doRun(const QString &s, const QStringList &args) {
	const bool ok = commonOpen(s, args);
	if(ok) {
		// update recent file list
		m_RecentFileManager->addFile(s);
	}
}

//------------------------------------------------------------------------------
// Name: doOpenFile(const QString &s)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::doOpenFile(const QString &s) {
	if(!s.isEmpty()) {
		m_LastOpenDirectory = QFileInfo(s).canonicalFilePath();
	
		doDetachFromProcess(false);
		
		const QStringList args = edb::v1::parseCommandLine(m_Arguments);
		doRun(s, args);
		
		edb::v1::analyzer()->invalidateAnalysis();
	}
}

//------------------------------------------------------------------------------
// Name: doAttach(edb::pid_t pid)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::doAttach(edb::pid_t pid) {
	// TODO: move this into debugger core
	if(pid == edb::v1::debuggerBase->pid()) {
		QMessageBox::information(this, tr("EDB"), tr("You are already debugging that process!"));
	} else {
		doDetachFromProcess(false);

		if(edb::v1::debuggerBase->attach(pid)) {
			edb::v1::analyzer()->invalidateAnalysis();
			
			m_WorkingDirectory = edb::v1::getProcessCurrentWorkingDirectory();

			setInitialDebuggerState();
			
			QStringList args = edb::v1::getProcessArguments();
			args.removeFirst();
			
			// make our arguments match what it was run with
			// TODO: this might be a mistake
			// perhaps we should have a seperate variable which represents 
			// 'actual arguments'
			m_Arguments = args.join(" ");
		} else {
			QMessageBox::information(this, tr("EDB"), tr("Failed to attach to process, please check priviledges and try again."));
		}

		edb::v1::debuggerBase->getState(edb::v1::currentState());
		updateGUI();
	}
}

//------------------------------------------------------------------------------
// Name: setDebuggerCaption(const QString &appname)
// Desc: sets the caption part to also show the application name and pid
//------------------------------------------------------------------------------
void DebuggerMain::setDebuggerCaption(const QString &appname) {
	setWindowTitle(QString(tr("EDB - %1 [%2]")).arg(appname).arg(edb::v1::debuggerBase->pid()));
}

//------------------------------------------------------------------------------
// Name: on_action_Open_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Open_triggered() {
	// TODO: move this into debugger core
	const QString filename = QFileDialog::getOpenFileName(this, tr("Choose a file"), m_LastOpenDirectory, QString());
	doOpenFile(filename);
}

//------------------------------------------------------------------------------
// Name: on_action_Attach_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Attach_triggered() {

	// TODO: the actual attach dialog should likely be provided by the Core plugin
	// since not all platforms/configurations will handle 'processes' in the same way

	static DialogAttach *const dlg = new DialogAttach(this);
	const int ret = dlg->exec();
	
	if(ret == QDialog::Accepted) {
		const QList<QTableWidgetItem *> sel = dlg->ui.tableWidget->selectedItems();
		if(sel.size() != 0) {
			const edb::pid_t pid = sel.begin()[0]->text().toUInt();
			doAttach(pid);
		}
	}
}

//------------------------------------------------------------------------------
// Name: on_action_Memory_Regions_triggered()
// Desc: displays the memory regions dialog, and optionally dumps some data
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Memory_Regions_triggered() {

	// TODO: the actual regions dialog should likely be provided by the Core plugin
	// since not all platforms/configurations will handle regions in the same way
	
	static DialogMemoryRegions *const dlg = new DialogMemoryRegions(this);	
	const int ret = dlg->exec();
	
	if(ret == QDialog::Accepted) {
		const MemoryRegions::Region *const region = dlg->selectedRegion();
		if(region != 0) {
			edb::v1::dumpData(region->start);
		}
	}
}

//------------------------------------------------------------------------------
// Name: mnuDumpCreateTab()
// Desc: duplicates the current tab creating a new one
//------------------------------------------------------------------------------
void DebuggerMain::mnuDumpCreateTab() {
	
	const int currentIndex = currentTab();
	
	// duplicate the current region
	DataViewInfo *const newDataView = new DataViewInfo(
		(currentIndex != -1) ? 
		m_DataRegions[currentIndex]->region : 
		MemoryRegions::Region()
		);
		
	QHexView *const hv = new QHexView;

	newDataView->view = hv;
		
	// setup the context menu
	hv->setContextMenuPolicy(Qt::CustomContextMenu);
	connect(hv, SIGNAL(customContextMenuRequested(const QPoint &)), SLOT(mnuDumpContextMenu(const QPoint &)));
	
	// show the initial data for this new view
	hv->setAddressOffset(newDataView->region.start);
	hv->setData(newDataView->stream);
	
	const Configuration &config = edb::v1::config();
		
	// set the default view options
	hv->setShowAddress(config.dataShowAddress);
	hv->setShowHexDump(config.dataShowHex);
	hv->setShowAsciiDump(config.dataShowAscii);
	hv->setShowComments(config.dataShowComments);
	hv->setRowWidth(config.dataRowWidth);
	hv->setWordWidth(config.dataWordWidth);
	
	// set the default font
	QFont dumpFont;
	dumpFont.fromString(config.dataFont);
	hv->setFont(dumpFont);

	m_DataRegions.push_back(newDataView);

	// create the tab!
	ui.tabWidget->addTab(hv, QString("%1-%2").arg(newDataView->region.start, EDB_MAX_HEX, 16, QChar('0')).arg(newDataView->region.end, EDB_MAX_HEX, 16, QChar('0')));

	ui.tabWidget->setCurrentIndex(ui.tabWidget->count() - 1);

	m_DelTab->setEnabled(ui.tabWidget->count() > 1);
}

//------------------------------------------------------------------------------
// Name: mnuDumpDeleteTab() 
// Desc: handles removing of a memory view tab
//------------------------------------------------------------------------------
void DebuggerMain::mnuDumpDeleteTab() {
	const int currentIndex = currentTab();
	
	// get a pointer to the info we need (before removing it from the list!)
	DataViewInfo *info = m_DataRegions[currentIndex];
		
	// remove it from the list
	m_DataRegions.remove(currentIndex);
	
	// remove the tab associated with it
	ui.tabWidget->removeTab(currentIndex);
	
	// update the delete button state
	m_DelTab->setEnabled(ui.tabWidget->count() > 1);
	
	// clean up the info object
	if(info != 0) {
		release(info->view);
	}
	
	release(info);
}

//------------------------------------------------------------------------------
// Name: doStepOver()
// Desc: attempts to step over an instruction, if it is determined that this can
//       be done, then it will place a one time breakpoint just after the
//       current instruction and run, otherwise it will just do an ordinary step
//------------------------------------------------------------------------------
template<void (DebuggerMain::*runFunc)(), void (DebuggerMain::*stepIntoFunc)()>
void DebuggerMain::doStepOver() {

	const edb::address_t ip = edb::v1::currentState().instructionPointer();

	quint8 buffer[edb::Instruction::MAX_SIZE];
	int sz = sizeof(buffer);

	if(edb::v1::getInstructionBytes(ip, buffer, sz)) {
		edb::Instruction insn(buffer, sz, 0, std::nothrow);
		if(insn.valid()) {
			if(edb::v1::archProcessor().canStepOver(insn)) {
				// add a temporary breakpoint at the instruction just 
				// after the call
				edb::v1::debuggerBase->addBreakpoint(ip + insn.size(), true);

				(this->*runFunc)();
				return;
			}
		}
	}

	// if all else fails, it's a step into
	(this->*stepIntoFunc)();
}

//------------------------------------------------------------------------------
// Name: currentTab() const
// Desc: 
//------------------------------------------------------------------------------
int DebuggerMain::currentTab() const {
	return ui.tabWidget->currentIndex();
}

//------------------------------------------------------------------------------
// Name: currentDataViewInfo() const
// Desc: 
//------------------------------------------------------------------------------
DebuggerMain::DataViewInfo *DebuggerMain::currentDataViewInfo() const {
	const int currentIndex = currentTab();
	return m_DataRegions[currentIndex];
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Plugins_triggered() {
	
	static QDialog *const dlg = new DialogPlugins(this);
	dlg->show();
}
